"""Scenic world model for Grand Theft Auto V (GTAV).

This interface must be used in `2D compatibility mode`.

Importing scenes generated using this model into GTA V requires a GTA V plugin,
which you can find `here <https://github.com/xyyue/scenic2gta>`__.
"""
